@@ -1,35 +1,32 @@
-import pendulum
 from typing import (
     Any,
     Callable,
     Dict,
     Iterable,
+    Iterator,
+    NamedTuple,
     Optional,
     Set,
-    Tuple,
-    Union,
-    List,
-    NamedTuple,
 )
+from contextlib import contextmanager
 
+import pendulum
 import prefect
-from prefect import config
 from prefect.core import Edge, Flow, Task
-from prefect.engine import signals
+from prefect.engine.result import Result
+from prefect.engine.results import ConstantResult
 from prefect.engine.runner import ENDRUN, Runner, call_state_handlers
 from prefect.engine.state import (
     Failed,
     Mapped,
     Pending,
-    Retrying,
     Running,
     Scheduled,
     State,
     Success,
 )
-from prefect.engine.task_runner import TaskRunner
+from prefect.utilities import executors
 from prefect.utilities.collections import flatten_seq
-from prefect.utilities.executors import run_with_heartbeat
 
 FlowRunnerInitializeResult = NamedTuple(
     "FlowRunnerInitializeResult",
@@ -60,7 +57,7 @@ class FlowRunner(Runner):
             opportunity to inspect or modify the new state. The handler
             will be passed the flow runner instance, the old (prior) state, and the new
             (current) state, with the following signature:
-                `state_handler(fr: FlowRunner, old_state: State, new_state: State) -> Optional[State]`
+            `state_handler(fr: FlowRunner, old_state: State, new_state: State) -> Optional[State]`
             If multiple functions are passed, then the `new_state` argument will be the
             result of the previous handler.
 
@@ -93,6 +90,9 @@ class FlowRunner(Runner):
         self.task_runner_cls = task_runner_cls
         super().__init__(state_handlers=state_handlers)
 
+    def __repr__(self) -> str:
+        return "<{}: {}>".format(type(self).__name__, self.flow.name)
+
     def call_runner_target_handlers(self, old_state: State, new_state: State) -> State:
         """
         A special state handler that the FlowRunner uses to call its flow's state handlers.
@@ -135,7 +135,8 @@ class FlowRunner(Runner):
             - task_states (Dict[Task, State]): a dictionary of any initial task states
             - context (Dict[str, Any], optional): prefect.Context to use for execution
                 to use for each Task run
-            - task_contexts (Dict[Task, Dict[str, Any]], optional): contexts that will be provided to each task
+            - task_contexts (Dict[Task, Dict[str, Any]], optional): contexts that will be
+                provided to each task
             - parameters(dict): the parameter values for the run
 
         Returns:
@@ -144,10 +145,12 @@ class FlowRunner(Runner):
         """
 
         # overwrite context parameters one-by-one
-        if parameters:
-            context_params = context.setdefault("parameters", {})
-            for param, value in parameters.items():
-                context_params[param] = value
+        context_params = context.setdefault("parameters", {})
+        for p in self.flow.parameters():
+            if not p.required:
+                context_params.setdefault(p.name, p.default)
+        for param, value in (parameters or {}).items():
+            context_params[param] = value
 
         context.update(flow_name=self.flow.name)
         context.setdefault("scheduled_start_time", pendulum.now("utc"))
@@ -168,8 +171,9 @@ class FlowRunner(Runner):
 
         for task in self.flow.tasks:
             task_contexts.setdefault(task, {}).update(
-                task_id=task.id, task_name=task.name, task_slug=task.slug
+                task_name=task.name, task_slug=self.flow.slugs[task]
             )
+
         state, context = super().initialize_run(state=state, context=context)
         return FlowRunnerInitializeResult(
             state=state,
@@ -185,7 +189,7 @@ class FlowRunner(Runner):
         return_tasks: Iterable[Task] = None,
         parameters: Dict[str, Any] = None,
         task_runner_state_handlers: Iterable[Callable] = None,
-        executor: "prefect.engine.executors.Executor" = None,
+        executor: "prefect.executors.Executor" = None,
         context: Dict[str, Any] = None,
         task_contexts: Dict[Task, Dict[str, Any]] = None,
     ) -> State:
@@ -204,19 +208,19 @@ class FlowRunner(Runner):
                 values, with keys being strings representing Parameter names and values being
                 their corresponding values
             - task_runner_state_handlers (Iterable[Callable], optional): A list of state change
-                handlers that will be provided to the task_runner, and called whenever a task changes
-                state.
+                handlers that will be provided to the task_runner, and called whenever a task
+                changes state.
             - executor (Executor, optional): executor to use when performing
                 computation; defaults to the executor specified in your prefect configuration
             - context (Dict[str, Any], optional): prefect.Context to use for execution
                 to use for each Task run
-            - task_contexts (Dict[Task, Dict[str, Any]], optional): contexts that will be provided to each task
+            - task_contexts (Dict[Task, Dict[str, Any]], optional): contexts that will be
+                provided to each task
 
         Returns:
             - State: `State` representing the final post-run state of the `Flow`.
 
         """
-
         self.logger.info("Beginning Flow run for '{}'".format(self.flow.name))
 
         # make copies to avoid modifying user inputs
@@ -225,7 +229,12 @@ class FlowRunner(Runner):
         task_contexts = dict(task_contexts or {})
         parameters = dict(parameters or {})
         if executor is None:
-            executor = prefect.engine.get_default_executor_class()()
+            # Use the executor on the flow, if configured
+            executor = getattr(self.flow, "executor", None)
+            if executor is None:
+                executor = prefect.engine.get_default_executor_class()()
+
+        self.logger.debug("Using executor type %s", type(executor).__name__)
 
         try:
             state, task_states, context, task_contexts = self.initialize_run(
@@ -254,18 +263,27 @@ class FlowRunner(Runner):
 
         # All other exceptions are trapped and turned into Failed states
         except Exception as exc:
-            self.logger.info(
+            self.logger.exception(
                 "Unexpected error while running flow: {}".format(repr(exc))
             )
             if prefect.context.get("raise_on_exception"):
                 raise exc
-            return Failed(
+            new_state = Failed(
                 message="Unexpected error while running flow: {}".format(repr(exc)),
                 result=exc,
             )
+            state = self.handle_state_change(state or Pending(), new_state)
 
         return state
 
+    @contextmanager
+    def check_for_cancellation(self) -> Iterator:
+        """Contextmanager used to wrap a cancellable section of a flow run.
+
+        No-op for the default `FlowRunner` class.
+        """
+        yield
+
     @call_state_handlers
     def check_flow_reached_start_time(self, state: State) -> State:
         """
@@ -334,14 +352,13 @@ class FlowRunner(Runner):
             - ENDRUN: if the flow is not pending or running
         """
         if state.is_pending():
-            self.logger.info("Starting flow run.")
             return Running(message="Running flow.")
         elif state.is_running():
             return state
         else:
             raise ENDRUN(state)
 
-    @run_with_heartbeat
+    @executors.run_with_heartbeat
     @call_state_handlers
     def get_flow_run_state(
         self,
@@ -350,7 +367,7 @@ class FlowRunner(Runner):
         task_contexts: Dict[Task, Dict[str, Any]],
         return_tasks: Set[Task],
         task_runner_state_handlers: Iterable[Callable],
-        executor: "prefect.engine.executors.base.Executor",
+        executor: "prefect.executors.base.Executor",
     ) -> State:
         """
         Runs the flow.
@@ -360,19 +377,25 @@ class FlowRunner(Runner):
                 `Pending`
             - task_states (dict): dictionary of task states to begin
                 computation with, with keys being Tasks and values their corresponding state
-            - task_contexts (Dict[Task, Dict[str, Any]]): contexts that will be provided to each task
+            - task_contexts (Dict[Task, Dict[str, Any]]): contexts that will be provided to
+                each task
             - return_tasks ([Task], optional): list of Tasks to include in the
                 final returned Flow state. Defaults to `None`
-            - task_runner_state_handlers (Iterable[Callable]): A list of state change
-                handlers that will be provided to the task_runner, and called whenever a task changes
+            - task_runner_state_handlers (Iterable[Callable]): A list of state change handlers
+                that will be provided to the task_runner, and called whenever a task changes
                 state.
-            - executor (Executor): executor to use when performing
-                computation; defaults to the executor provided in your prefect configuration
+            - executor (Executor): executor to use when performing computation; defaults to the
+                executor provided in your prefect configuration
 
         Returns:
             - State: `State` representing the final post-run state of the `Flow`.
 
         """
+        # this dictionary is used for tracking the states of "children" mapped tasks;
+        # when running on Dask, we want to avoid serializing futures, so instead
+        # of storing child task states in the `map_states` attribute we instead store
+        # in this dictionary and only after they are resolved do we attach them to the Mapped state
+        mapped_children = dict()  # type: Dict[Task, list]
 
         if not state.is_running():
             self.logger.info("Flow is not in a Running state.")
@@ -383,41 +406,202 @@ class FlowRunner(Runner):
         if set(return_tasks).difference(self.flow.tasks):
             raise ValueError("Some tasks in return_tasks were not found in the flow.")
 
+        def extra_context(task: Task, task_index: int = None) -> dict:
+            return {
+                "task_name": task.name,
+                "task_tags": task.tags,
+                "task_index": task_index,
+            }
+
         # -- process each task in order
 
-        with executor.start():
+        with self.check_for_cancellation(), executor.start():
 
             for task in self.flow.sorted_tasks():
-
                 task_state = task_states.get(task)
 
-                # if the state is finished, don't run the task, just use the provided state
+                # if a task is a constant task, we already know its return value
+                # no need to use up resources by running it through a task runner
+                if task_state is None and isinstance(
+                    task, prefect.tasks.core.constants.Constant
+                ):
+                    task_states[task] = task_state = Success(result=task.value)
+
+                # Always restart completed resource setup/cleanup tasks unless
+                # they were explicitly cached.
+                # TODO: we only need to rerun these tasks if any pending
+                # downstream tasks depend on them.
+                if (
+                    isinstance(
+                        task,
+                        (
+                            prefect.tasks.core.resource_manager.ResourceSetupTask,
+                            prefect.tasks.core.resource_manager.ResourceCleanupTask,
+                        ),
+                    )
+                    and task_state is not None
+                    and task_state.is_finished()
+                    and not task_state.is_cached()
+                ):
+                    task_states[task] = task_state = Pending()
+
+                # if the state is finished, don't run the task, just use the provided state if
+                # the state is cached / mapped, we still want to run the task runner pipeline
+                # steps to either ensure the cache is still valid / or to recreate the mapped
+                # pipeline for possible retries
                 if (
                     isinstance(task_state, State)
                     and task_state.is_finished()
+                    and not task_state.is_cached()
                     and not task_state.is_mapped()
                 ):
                     continue
 
-                upstream_states = {}  # type: Dict[Edge, Union[State, Iterable]]
+                upstream_states = {}  # type: Dict[Edge, State]
+
+                # this dictionary is used exclusively for "reduce" tasks in particular we store
+                # the states / futures corresponding to the upstream children, and if running
+                # on Dask, let Dask resolve them at the appropriate time.
+                # Note: this is an optimization that allows Dask to resolve the mapped
+                # dependencies by "elevating" them to a function argument.
+                upstream_mapped_states = {}  # type: Dict[Edge, list]
 
                 # -- process each edge to the task
                 for edge in self.flow.edges_to(task):
+
+                    # load the upstream task states (supplying Pending as a default)
                     upstream_states[edge] = task_states.get(
                         edge.upstream_task, Pending(message="Task state not available.")
                     )
 
-                # -- run the task
+                    # if the edge is flattened and not the result of a map, then we
+                    # preprocess the upstream states. If it IS the result of a
+                    # map, it will be handled in `prepare_upstream_states_for_mapping`
+                    if edge.flattened:
+                        if not isinstance(upstream_states[edge], Mapped):
+                            upstream_states[edge] = executor.submit(
+                                executors.flatten_upstream_state, upstream_states[edge]
+                            )
 
-                task_states[task] = executor.submit(
-                    self.run_task,
-                    task=task,
-                    state=task_state,
-                    upstream_states=upstream_states,
-                    context=dict(prefect.context, **task_contexts.get(task, {})),
-                    task_runner_state_handlers=task_runner_state_handlers,
-                    executor=executor,
-                )
+                    # this checks whether the task is a "reduce" task for a mapped pipeline
+                    # and if so, collects the appropriate upstream children
+                    if not edge.mapped and isinstance(upstream_states[edge], Mapped):
+                        children = mapped_children.get(edge.upstream_task, [])
+
+                        # if the edge is flattened, then we need to wait for the mapped children
+                        # to complete and then flatten them
+                        if edge.flattened:
+                            children = executors.flatten_mapped_children(
+                                mapped_children=children, executor=executor
+                            )
+
+                        upstream_mapped_states[edge] = children
+
+                # augment edges with upstream constants
+                for key, val in self.flow.constants[task].items():
+                    edge = Edge(
+                        upstream_task=prefect.tasks.core.constants.Constant(val),
+                        downstream_task=task,
+                        key=key,
+                    )
+                    upstream_states[edge] = Success(
+                        "Auto-generated constant value",
+                        result=ConstantResult(value=val),
+                    )
+
+                # handle mapped tasks
+                if any(edge.mapped for edge in upstream_states.keys()):
+
+                    # wait on upstream states to determine the width of the pipeline
+                    # this is the key to depth-first execution
+                    upstream_states = executor.wait(
+                        {e: state for e, state in upstream_states.items()}
+                    )
+                    # we submit the task to the task runner to determine if
+                    # we can proceed with mapping - if the new task state is not a Mapped
+                    # state then we don't proceed
+                    task_states[task] = executor.wait(
+                        executor.submit(
+                            run_task,
+                            task=task,
+                            state=task_state,  # original state
+                            upstream_states=upstream_states,
+                            context=dict(
+                                prefect.context, **task_contexts.get(task, {})
+                            ),
+                            flow_result=self.flow.result,
+                            task_runner_cls=self.task_runner_cls,
+                            task_runner_state_handlers=task_runner_state_handlers,
+                            upstream_mapped_states=upstream_mapped_states,
+                            is_mapped_parent=True,
+                            extra_context=extra_context(task),
+                        )
+                    )
+
+                    # either way, we should now have enough resolved states to restructure
+                    # the upstream states into a list of upstream state dictionaries to iterate over
+                    list_of_upstream_states = (
+                        executors.prepare_upstream_states_for_mapping(
+                            task_states[task],
+                            upstream_states,
+                            mapped_children,
+                            executor=executor,
+                        )
+                    )
+
+                    submitted_states = []
+
+                    for idx, states in enumerate(list_of_upstream_states):
+                        # if we are on a future rerun of a partially complete flow run,
+                        # there might be mapped children in a retrying state; this check
+                        # looks into the current task state's map_states for such info
+                        if (
+                            isinstance(task_state, Mapped)
+                            and len(task_state.map_states) >= idx + 1
+                        ):
+                            current_state = task_state.map_states[
+                                idx
+                            ]  # type: Optional[State]
+                        elif isinstance(task_state, Mapped):
+                            current_state = None
+                        else:
+                            current_state = task_state
+
+                        # this is where each child is submitted for actual work
+                        submitted_states.append(
+                            executor.submit(
+                                run_task,
+                                task=task,
+                                state=current_state,
+                                upstream_states=states,
+                                context=dict(
+                                    prefect.context,
+                                    **task_contexts.get(task, {}),
+                                    map_index=idx,
+                                ),
+                                flow_result=self.flow.result,
+                                task_runner_cls=self.task_runner_cls,
+                                task_runner_state_handlers=task_runner_state_handlers,
+                                upstream_mapped_states=upstream_mapped_states,
+                                extra_context=extra_context(task, task_index=idx),
+                            )
+                        )
+                    if isinstance(task_states.get(task), Mapped):
+                        mapped_children[task] = submitted_states  # type: ignore
+
+                else:
+                    task_states[task] = executor.submit(
+                        run_task,
+                        task=task,
+                        state=task_state,
+                        upstream_states=upstream_states,
+                        context=dict(prefect.context, **task_contexts.get(task, {})),
+                        flow_result=self.flow.result,
+                        task_runner_cls=self.task_runner_cls,
+                        task_runner_state_handlers=task_runner_state_handlers,
+                        upstream_mapped_states=upstream_mapped_states,
+                        extra_context=extra_context(task),
+                    )
 
             # ---------------------------------------------
             # Collect results
@@ -443,7 +627,9 @@ class FlowRunner(Runner):
             all_final_states = final_states.copy()
             for t, s in list(final_states.items()):
                 if s.is_mapped():
-                    s.map_states = executor.wait(s.map_states)
+                    # ensure we wait for any mapped children to complete
+                    if t in mapped_children:
+                        s.map_states = executor.wait(mapped_children[t])
                     s.result = [ms.result for ms in s.map_states]
                     all_final_states[t] = s.map_states
 
@@ -456,6 +642,7 @@ class FlowRunner(Runner):
         return_states = {t: final_states[t] for t in return_tasks}
 
         state = self.determine_final_state(
+            state=state,
             key_states=key_states,
             return_states=return_states,
             terminal_states=terminal_states,
@@ -465,6 +652,7 @@ class FlowRunner(Runner):
 
     def determine_final_state(
         self,
+        state: State,
         key_states: Set[State],
         return_states: Dict[Task, State],
         terminal_states: Set[State],
@@ -473,19 +661,19 @@ class FlowRunner(Runner):
         Implements the logic for determining the final state of the flow run.
 
         Args:
-            - key_states (Set[State]): the states which will determine the success / failure of the flow run
+            - state (State): the current state of the Flow
+            - key_states (Set[State]): the states which will determine the success / failure of
+                the flow run
             - return_states (Dict[Task, State]): states to return as results
             - terminal_states (Set[State]): the states of the terminal tasks for this flow
 
         Returns:
             - State: the final state of the flow run
         """
-        state = State()  # mypy initialization
-
         # check that the flow is finished
         if not all(s.is_finished() for s in terminal_states):
             self.logger.info("Flow run RUNNING: terminal tasks are incomplete.")
-            state = Running(message="Flow run in progress.", result=return_states)
+            state.result = return_states
 
         # check if any key task failed
         elif any(s.is_failed() for s in key_states):
@@ -506,55 +694,57 @@ class FlowRunner(Runner):
 
         return state
 
-    def run_task(
-        self,
-        task: Task,
-        state: State,
-        upstream_states: Dict[Edge, State],
-        context: Dict[str, Any],
-        task_runner_state_handlers: Iterable[Callable],
-        executor: "prefect.engine.executors.Executor",
-    ) -> State:
-        """
 
-        Runs a specific task. This method is intended to be called by submitting it to
-        an executor.
+def run_task(
+    task: Task,
+    state: State,
+    upstream_states: Dict[Edge, State],
+    context: Dict[str, Any],
+    flow_result: Result,
+    task_runner_cls: Callable,
+    task_runner_state_handlers: Iterable[Callable],
+    upstream_mapped_states: Dict[Edge, list],
+    is_mapped_parent: bool = False,
+) -> State:
+    """
+    Runs a specific task. This method is intended to be called by submitting it to
+    an executor.
 
-        Args:
-            - task (Task): the task to run
-            - state (State): starting state for the Flow. Defaults to
-                `Pending`
-            - upstream_states (Dict[Edge, State]): dictionary of upstream states
-            - context (Dict[str, Any]): a context dictionary for the task run
-            - task_runner_state_handlers (Iterable[Callable]): A list of state change
-                handlers that will be provided to the task_runner, and called whenever a task changes
-                state.
-            - executor (Executor): executor to use when performing
-                computation; defaults to the executor provided in your prefect configuration
+    Args:
+        - task (Task): the task to run
+        - state (State): starting state for the Flow. Defaults to `Pending`
+        - task_runner_cls (Callable): the `TaskRunner` class to use
+        - upstream_states (Dict[Edge, State]): dictionary of upstream states
+        - context (Dict[str, Any]): a context dictionary for the task run
+        - flow_result (Result): the `Result` associated with the flow (if any)
+        - task_runner_state_handlers (Iterable[Callable]): A list of state change
+            handlers that will be provided to the task_runner, and called
+            whenever a task changes state.
+        - upstream_mapped_states (Dict[Edge, list]): dictionary of upstream states
+            corresponding to mapped children dependencies
+        - is_mapped_parent (bool): a boolean indicating whether this task run is the
+            run of a parent mapped task
 
-        Returns:
-            - State: `State` representing the final post-run state of the `Flow`.
-
-        """
-        default_handler = task.result_handler or self.flow.result_handler
-        task_runner = self.task_runner_cls(
-            task=task,
-            state_handlers=task_runner_state_handlers,
-            result_handler=default_handler,
-        )
-
-        # if this task reduces over a mapped state, make sure its children have finished
+    Returns:
+        - State: `State` representing the final post-run state of the `Flow`.
+    """
+    with prefect.context(context):
+        # Update upstream_states with info from upstream_mapped_states
         for edge, upstream_state in upstream_states.items():
-
-            # if the upstream state is Mapped, wait until its results are all available
             if not edge.mapped and upstream_state.is_mapped():
                 assert isinstance(upstream_state, Mapped)  # mypy assert
-                upstream_state.map_states = executor.wait(upstream_state.map_states)
+                upstream_state.map_states = upstream_mapped_states.get(
+                    edge, upstream_state.map_states
+                )
                 upstream_state.result = [s.result for s in upstream_state.map_states]
-
+        task_runner = task_runner_cls(
+            task=task,
+            state_handlers=task_runner_state_handlers,
+            flow_result=flow_result,
+        )
         return task_runner.run(
             state=state,
             upstream_states=upstream_states,
+            is_mapped_parent=is_mapped_parent,
             context=context,
-            executor=executor,
         )
