@@ -9,7 +9,8 @@ class EdgeSchema(ObjectSchema):
     class Meta:
         object_class = lambda: prefect.core.Edge
 
-    upstream_task = fields.Nested(TaskSchema, only=["id"])
-    downstream_task = fields.Nested(TaskSchema, only=["id"])
+    upstream_task = fields.Nested(TaskSchema, only=["slug"])
+    downstream_task = fields.Nested(TaskSchema, only=["slug"])
     key = fields.String(allow_none=True)
     mapped = fields.Boolean(allow_none=True)
+    flattened = fields.Boolean(allow_none=True)
