@@ -1,10 +1,12 @@
-from typing import Any
+import time
+from typing import Any, cast
 
-from kubernetes import client, config
+from kubernetes import client
 
 from prefect import Task
-from prefect.client import Secret
+from prefect.engine import signals
 from prefect.utilities.tasks import defaults_from_attrs
+from prefect.utilities.kubernetes import get_kubernetes_client
 
 
 class CreateNamespacedJob(Task):
@@ -16,7 +18,10 @@ class CreateNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -46,7 +51,7 @@ class CreateNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.body = body or {}
         self.namespace = namespace
@@ -85,19 +90,9 @@ class CreateNamespacedJob(Task):
         if not body:
             raise ValueError("A dictionary representing a V1Job must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
@@ -114,7 +109,10 @@ class DeleteNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -143,7 +141,7 @@ class DeleteNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.job_name = job_name
         self.namespace = namespace
@@ -161,6 +159,7 @@ class DeleteNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
+        delete_option_kwargs: dict = None,
     ) -> None:
         """
         Task run method.
@@ -174,6 +173,9 @@ class DeleteNamespacedJob(Task):
             - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                 which stored your Kubernetes API Key; this Secret must be a string and in
                 BearerToken format
+            - delete_option_kwargs (dict, optional): Optional keyword arguments to pass to
+                the V1DeleteOptions object (e.g. {"propagation_policy": "...",
+                "grace_period_seconds": "..."}.
 
         Raises:
             - ValueError: if `job_name` is `None`
@@ -181,27 +183,18 @@ class DeleteNamespacedJob(Task):
         if not job_name:
             raise ValueError("The name of a Kubernetes job must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
+        delete_option_kwargs = delete_option_kwargs or {}
 
         api_client.delete_namespaced_job(
             name=job_name,
             namespace=namespace,
-            body=api_client.V1DeleteOptions(),
-            **kube_kwargs
+            body=client.V1DeleteOptions(**delete_option_kwargs),
+            **kube_kwargs,
         )
 
 
@@ -214,7 +207,10 @@ class ListNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -241,7 +237,7 @@ class ListNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.namespace = namespace
         self.kube_kwargs = kube_kwargs or {}
@@ -268,22 +264,12 @@ class ListNamespacedJob(Task):
                 which stored your Kubernetes API Key; this Secret must be a string and in
                 BearerToken format
 
-        Return:
+        Returns:
             - V1JobList: a Kubernetes V1JobList of the jobs which are found
         """
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
 
@@ -299,7 +285,10 @@ class PatchNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -331,7 +320,7 @@ class PatchNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.job_name = job_name
         self.body = body or {}
@@ -379,19 +368,9 @@ class PatchNamespacedJob(Task):
         if not job_name:
             raise ValueError("The name of a Kubernetes job must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
@@ -410,7 +389,10 @@ class ReadNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -439,7 +421,7 @@ class ReadNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.job_name = job_name
         self.namespace = namespace
@@ -480,19 +462,9 @@ class ReadNamespacedJob(Task):
         if not job_name:
             raise ValueError("The name of a Kubernetes job must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
 
@@ -510,7 +482,10 @@ class ReplaceNamespacedJob(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -542,7 +517,7 @@ class ReplaceNamespacedJob(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
-        **kwargs: Any
+        **kwargs: Any,
     ):
         self.job_name = job_name
         self.body = body or {}
@@ -588,19 +563,9 @@ class ReplaceNamespacedJob(Task):
         if not job_name:
             raise ValueError("The name of a Kubernetes job must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.BatchV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.BatchV1Api()
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
@@ -608,3 +573,156 @@ class ReplaceNamespacedJob(Task):
         api_client.replace_namespaced_job(
             name=job_name, namespace=namespace, body=body, **kube_kwargs
         )
+
+
+class RunNamespacedJob(Task):
+    """
+    Task for running a namespaced job on Kubernetes.
+    This task first creates a job on a Kubernetes cluster according to the specification
+    given in the body, and then by default regularly checks its status at 5-second intervals.
+    After the job is successfully completed, all resources by default are deleted: job and the
+    corresponding pods. If job is in the failed status, resources will not be removed
+    from the cluster so that user can check the logs on the cluster.
+
+    Note that all initialization arguments can optionally be provided or overwritten at runtime.
+
+    This task will attempt to connect to a Kubernetes cluster in three steps with
+    the first successful connection attempt becoming the mode of communication with a
+    cluster.
+
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
+    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
+    3. Attempt out-of-cluster connection using the default location for a kube config file
+
+    The arguments `body` and `kube_kwargs` will perform an in-place update when the task
+    is run. This means that it is possible to provide `body = {"info": "here"}` at
+    instantiation and then provide `body = {"more": "info"}` at run time which will make
+    `body = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
+    and runtime will be replaced with the runtime value.
+
+    Args:
+        - body (dict, optional): A dictionary representation of a Kubernetes V1Job
+            specification
+        - namespace (str, optional): The Kubernetes namespace to create this job in,
+            defaults to the `default` namespace
+        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
+            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
+        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
+            which stored your Kubernetes API Key; this Secret must be a string and in
+            BearerToken format
+        - job_status_pool_interval (int, optional): The interval given in seconds
+            indicating how often the Kubernetes API will be requested about the status
+            of the job being performed, defaults to the `5` seconds
+        - delete_job_after_completion (bool, optional): boolean value determining whether
+            resources related to a given job will be removed from the Kubernetes cluster
+            after completion, defaults to the `True` value
+        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
+            constructor
+    """
+
+    def __init__(
+        self,
+        body: dict = None,
+        namespace: str = "default",
+        kube_kwargs: dict = None,
+        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
+        job_status_pool_interval: int = 5,
+        delete_job_after_completion: bool = True,
+        **kwargs: Any,
+    ):
+        self.body = body or {}
+        self.namespace = namespace
+        self.kube_kwargs = kube_kwargs or {}
+        self.kubernetes_api_key_secret = kubernetes_api_key_secret
+        self.job_status_pool_interval = job_status_pool_interval
+        self.delete_job_after_completion = delete_job_after_completion
+
+        super().__init__(**kwargs)
+
+    @defaults_from_attrs(
+        "body",
+        "namespace",
+        "kube_kwargs",
+        "kubernetes_api_key_secret",
+        "job_status_pool_interval",
+        "delete_job_after_completion",
+    )
+    def run(
+        self,
+        body: dict = None,
+        namespace: str = "default",
+        kube_kwargs: dict = None,
+        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
+        job_status_pool_interval: int = 5,
+        delete_job_after_completion: bool = True,
+    ) -> None:
+        """
+        Task run method.
+
+        Args:
+            - body (dict, optional): A dictionary representation of a Kubernetes V1Job
+                specification
+            - namespace (str, optional): The Kubernetes namespace to create this job in,
+                defaults to the `default` namespace
+            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
+                Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
+            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
+                which stored your Kubernetes API Key; this Secret must be a string and in
+                BearerToken format
+            - job_status_pool_interval (int, optional): The interval given in seconds
+                indicating how often the Kubernetes API will be requested about the status
+                of the job being performed, defaults to the `5` seconds
+            - delete_job_after_completion (bool, optional): boolean value determining whether
+                resources related to a given job will be removed from the Kubernetes cluster
+                after completion, defaults to the `True` value
+
+        Raises:
+            - ValueError: if `body` is `None`
+            - ValueError: if `body["metadata"]["name"] is `None`
+        """
+        if not body:
+            raise ValueError("A dictionary representing a V1Job must be provided.")
+
+        body = {**self.body, **(body or {})}
+        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
+
+        job_name = body.get("metadata", {}).get("name")
+        if not job_name:
+            raise ValueError(
+                "The job name must be defined in the body under the metadata key."
+            )
+
+        api_client = cast(
+            client.BatchV1Api, get_kubernetes_client("job", kubernetes_api_key_secret)
+        )
+
+        api_client.create_namespaced_job(namespace=namespace, body=body, **kube_kwargs)
+        self.logger.info(f"Job {job_name} has been created.")
+
+        completed = False
+        while not completed:
+            res = api_client.read_namespaced_job_status(
+                name=job_name, namespace=namespace
+            )
+
+            if res.status.active:
+                time.sleep(job_status_pool_interval)
+            else:
+                if res.status.failed:
+                    raise signals.FAIL(
+                        f"Job {job_name} failed, check Kubernetes pod logs for more information."
+                    )
+                elif res.status.succeeded:
+                    self.logger.info(f"Job {job_name} has been completed.")
+                    completed = True
+
+        if delete_job_after_completion:
+            api_client.delete_namespaced_job(
+                name=job_name,
+                namespace=namespace,
+                body=client.V1DeleteOptions(propagation_policy="Foreground"),
+            )
+            self.logger.info(f"Job {job_name} has been deleted.")
