@@ -1,10 +1,10 @@
-from typing import Any
+from typing import Any, cast
 
-from kubernetes import client, config
+from kubernetes import client
 
 from prefect import Task
-from prefect.client import Secret
 from prefect.utilities.tasks import defaults_from_attrs
+from prefect.utilities.kubernetes import get_kubernetes_client
 
 
 class CreateNamespacedPod(Task):
@@ -16,7 +16,10 @@ class CreateNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -85,19 +88,9 @@ class CreateNamespacedPod(Task):
         if not body:
             raise ValueError("A dictionary representing a V1Pod must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
@@ -114,7 +107,10 @@ class DeleteNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -161,6 +157,7 @@ class DeleteNamespacedPod(Task):
         namespace: str = "default",
         kube_kwargs: dict = None,
         kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
+        delete_option_kwargs: dict = None,
     ) -> None:
         """
         Task run method.
@@ -174,6 +171,9 @@ class DeleteNamespacedPod(Task):
             - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                 which stored your Kubernetes API Key; this Secret must be a string and in
                 BearerToken format
+            - delete_option_kwargs (dict, optional): Optional keyword arguments to pass to
+                the V1DeleteOptions object (e.g. {"propagation_policy": "...",
+                "grace_period_seconds": "..."}.
 
         Raises:
             - ValueError: if `pod_name` is `None`
@@ -181,26 +181,17 @@ class DeleteNamespacedPod(Task):
         if not pod_name:
             raise ValueError("The name of a Kubernetes pod must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
+        delete_option_kwargs = delete_option_kwargs or {}
 
         api_client.delete_namespaced_pod(
             name=pod_name,
             namespace=namespace,
-            body=api_client.V1DeleteOptions(),
+            body=client.V1DeleteOptions(**delete_option_kwargs),
             **kube_kwargs
         )
 
@@ -214,7 +205,10 @@ class ListNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -268,22 +262,12 @@ class ListNamespacedPod(Task):
                 which stored your Kubernetes API Key; this Secret must be a string and in
                 BearerToken format
 
-        Return:
+        Returns:
             - V1PodList: a Kubernetes V1PodList of the pods which are found
         """
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
 
@@ -299,7 +283,10 @@ class PatchNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -379,19 +366,9 @@ class PatchNamespacedPod(Task):
         if not pod_name:
             raise ValueError("The name of a Kubernetes pod must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
@@ -410,7 +387,10 @@ class ReadNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -480,19 +460,9 @@ class ReadNamespacedPod(Task):
         if not pod_name:
             raise ValueError("The name of a Kubernetes pod must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
 
@@ -510,7 +480,10 @@ class ReplaceNamespacedPod(Task):
     the first successful connection attempt becoming the mode of communication with a
     cluster.
 
-    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key
+    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
+    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
+    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
+    an override for the remote connection.
     2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
     3. Attempt out-of-cluster connection using the default location for a kube config file
 
@@ -588,19 +561,9 @@ class ReplaceNamespacedPod(Task):
         if not pod_name:
             raise ValueError("The name of a Kubernetes pod must be provided.")
 
-        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()
-
-        if kubernetes_api_key:
-            configuration = client.Configuration()
-            configuration.api_key["authorization"] = kubernetes_api_key
-            api_client = client.CoreV1Api(client.ApiClient(configuration))
-        else:
-            try:
-                config.load_incluster_config()
-            except config.config_exception.ConfigException:
-                config.load_kube_config()
-
-            api_client = client.CoreV1Api()
+        api_client = cast(
+            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
+        )
 
         body = {**self.body, **(body or {})}
         kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
