@@ -1,10 +1,13 @@
 import prefect.utilities.logging
+import prefect.utilities.backend
 import prefect.utilities.collections
 import prefect.utilities.configuration
 import prefect.utilities.datetimes
+import prefect.utilities.diagnostics
 import prefect.utilities.exceptions
-import prefect.utilities.environments
 import prefect.utilities.graphql
 import prefect.utilities.notifications
 import prefect.utilities.serialization
 import prefect.utilities.tasks
+import prefect.utilities.agent
+import prefect.utilities.storage
